#define NOCRT

#include <windows.h>
#ifndef NOCRT
#include <math.h>
#endif
#include "resource.h"

//#define WIDTH 300
//#define HEIGHT 300
unsigned int WIDTH=300;
unsigned int HEIGHT=300;

HWND hMainWnd;

int* field1 = NULL;
int* field2 = NULL;
int* field3 = NULL;

// DefaultValues
unsigned int startSize = 4;
unsigned int startHeight = 1024;

int numerator = 63;
int denominator = 64;

BOOL movePixel = TRUE;
BOOL showHighlight = TRUE;
BOOL showShadow = FALSE;

BOOL keepPress = FALSE;
BOOL moveRaise = TRUE;

/**************************************/
/*       C R T  S E C T I O N         */
/**************************************/
#ifdef NOCRT
int _fltused = 0xC0;	// damasu. 0xC0 wa nandemo ok.

#undef ZeroMemory
void ZeroMemory(void* ptr, DWORD size)
{
	while(size--)
		*((char*)ptr)++ = 0;
}

void* malloc(size_t size)
{
	return HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, size);
}

void free(void* ptr)
{
	HeapFree(GetProcessHeap(), 0, ptr);
}

#ifdef _DEBUG	// Why RELEASE BUILD can't parse sqrt/cos?
double sqrt(double a)
{
	int i;
	double result;

	for(result = 1; result * result < a; result++)
		;

	for(i = 0; i < 10; i++)
		result = (result + a / result) / 2;

	return result;
}

double cos(double a)
{
	return 1 - a*a/2 + a*a*a*a/24;// - a*a*a*a*a*a/720 + a*a*a*a*a*a*a*a/40320;
}
#endif	// _DEBUG
#endif	// NOCRT

/**************************************/
/*       E N D  S E C T I O N         */
/**************************************/

static signed int shiftx = 0, shifty = 0;

void initwave(void)
{
	unsigned int i, x, y;

	for(i = 0, y = 0; y < HEIGHT; y++)
		for(x = 0; x < WIDTH; x++, i++)
			field1[i] = field2[i] = field3[i] = 128;

	return;
}

void dowave(DWORD* bits, char* backfile)
{
	unsigned int i;
	unsigned int iMax = WIDTH * (HEIGHT - 1) - 1;
	DWORD* backbits;
	DWORD backwidth;
	DWORD backheight;

	if(backfile)
	{
		backbits = (DWORD*)(backfile + ((BITMAPFILEHEADER*)backfile)->bfOffBits);
		backwidth = ((BITMAPINFOHEADER*)(backfile + sizeof(BITMAPFILEHEADER)))->biWidth;
		backheight = ((BITMAPINFOHEADER*)(backfile + sizeof(BITMAPFILEHEADER)))->biHeight;
	}

	for(i = WIDTH + 1; i < iMax; i++)
	{
		unsigned int x = i % WIDTH;
		unsigned int y = i / WIDTH;
		int a;
		int r, g, b;

		if(x == 0 || x == WIDTH - 1)
			continue;

		// g
		//field3[i] = (((((int)field2[i - 1] + (int)field2[i + 1] + (int)field2[i - WIDTH] + (int)field2[i + WIDTH]) >> 1) - (int)field1[i]) * 63) >> 6;
		field3[i] = (((field2[i - 1] + field2[i + 1] + field2[i - WIDTH] + field2[i + WIDTH]) >> 1) - field1[i]) * numerator / denominator;
		//a = field2[i] / 64 + 128;
		a = ((field2[i] - field2[i - WIDTH - 1]) >> 8);// + 128;

		if(a > 255)
			a = 255;
		if(a < 0)
			a = 0;

		//bits[i] = RGB(255, 192 + (a / 4), 128 + (a / 2));
		r = g = b = 0;
		if(backfile)
		{
			unsigned int xx;
			unsigned int yy;

			if(movePixel)
			{
				xx = x + ((signed int)a - 128) / 16 + 128/16 + shiftx;
				yy = y + ((signed int)a - 128) / 16 + 128/16 + shifty;
			}else
			{
				xx = x + shiftx;
				yy = y + shifty;
			}
			/*
			if(xx < 0)
				xx = 0;
			if(backwidth <= xx)
				xx = backwidth - 1;
			if(yy < 0)
				yy = 0;
			if(backheight <= yy)
				yy = backheight - 1;
			*/
			if(0 <= xx && 0 <= yy && xx < backwidth && yy < backheight)
			{
				int debug = (backheight - yy) * backwidth + xx;
				DWORD back = *(DWORD*)((char*)backbits + debug*3);

				//r = (*(RGBQUAD*)&back).rgbRed; + a;
				//g = (*(RGBQUAD*)&back).rgbGreen; + a;
				//b = (*(RGBQUAD*)&back).rgbBlue; + a;
				b = (*(RGBTRIPLE*)&back).rgbtRed;
				g = (*(RGBTRIPLE*)&back).rgbtGreen;
				r = (*(RGBTRIPLE*)&back).rgbtBlue;
			}
		}

		if(showHighlight)
		{
			b += a;
			g += a;
			r += a;
		}
		if(showShadow)	// oops, it is bug yet.
		{
			b = b + (a - 128) / 2;
			g = g + (a - 128) / 2;
			r = r + (a - 128) / 2;
		}

		if(r < 0)
			r = 0;
		if(g < 0)
			g = 0;
		if(b < 0)
			b = 0;
		if(r > 255)
			r = 255;
		if(g > 255)
			g = 255;
		if(b > 255)
			b = 255;

		bits[i] = RGB(r, g, b);
	}

	{
		int* field0;

		field0 = field1;
		field1 = field2;
		field2 = field3;
		field3 = field0;
	}

	return;
}

void raisewave(int* field, unsigned int cx, unsigned int cy)
{
	signed int x, y;

	if (cx < startSize || cx > WIDTH - 1 - startSize || cy < startSize || cy > HEIGHT - 1 - startSize)
		return;

	for (x = -(signed int)startSize; x < (signed)startSize; x++)
		for (y = -(signed int)startSize; y < (signed)startSize; y++)
		{
			double r = (double)x * x + y * y;
			int ad = (x + cx) + (y + cy) * WIDTH;

			if (r < startSize * startSize)
			{
				field[ad] += (int)((cos(sqrt(r) / startSize * 3.14159) + 1) * 256 * startHeight);
				// if (field[ad] > 256 * 120) field[ad] = 256 * 120;
			}
		}

	return;
}

BOOL CALLBACK ConfigureWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		SetDlgItemInt(hWnd, IDC_STARTSIZE, startSize, FALSE);
		SetDlgItemInt(hWnd, IDC_STARTHEIGHT, startHeight, FALSE);

		SetDlgItemInt(hWnd, IDC_NUMERATOR, numerator, FALSE);
		SetDlgItemInt(hWnd, IDC_DENOMINATOR, denominator, FALSE);

		SendMessage(GetDlgItem(hWnd, IDC_MOVEPIXEL), BM_SETCHECK, movePixel ? BST_CHECKED : 0, 0);
		SendMessage(GetDlgItem(hWnd, IDC_SHOWHIGHLIGHT), BM_SETCHECK, showHighlight ? BST_CHECKED : 0, 0);
		SendMessage(GetDlgItem(hWnd, IDC_SHOWSHADOW), BM_SETCHECK, showShadow ? BST_CHECKED : 0, 0);
		SendMessage(GetDlgItem(hWnd, IDC_MOVERAISE), BM_SETCHECK, moveRaise ? BST_CHECKED : 0, 0);
		SendMessage(GetDlgItem(hWnd, IDC_KEEPPRESS), BM_SETCHECK, keepPress ? BST_CHECKED : 0, 0);

		EnableWindow(GetDlgItem(hWnd, IDC_KEEPPRESS), moveRaise);

		return TRUE;	// Hey Windoze, Please set focus to default item!! PLEASE!!
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			startSize = GetDlgItemInt(hWnd, IDC_STARTSIZE, NULL, FALSE);
			startHeight = GetDlgItemInt(hWnd, IDC_STARTHEIGHT, NULL, FALSE);

			numerator = GetDlgItemInt(hWnd, IDC_NUMERATOR, NULL, FALSE);
			denominator = GetDlgItemInt(hWnd, IDC_DENOMINATOR, NULL, FALSE);

			movePixel = SendMessage(GetDlgItem(hWnd, IDC_MOVEPIXEL), BM_GETCHECK, 0, 0) & BST_CHECKED ? TRUE : FALSE;
			showHighlight = SendMessage(GetDlgItem(hWnd, IDC_SHOWHIGHLIGHT), BM_GETCHECK, 0, 0) & BST_CHECKED ? TRUE : FALSE;
			showShadow = SendMessage(GetDlgItem(hWnd, IDC_SHOWSHADOW), BM_GETCHECK, 0, 0) & BST_CHECKED ? TRUE : FALSE;
			moveRaise = SendMessage(GetDlgItem(hWnd, IDC_MOVERAISE), BM_GETCHECK, 0, 0) & BST_CHECKED ? TRUE : FALSE;
			if(moveRaise)
				keepPress = SendMessage(GetDlgItem(hWnd, IDC_KEEPPRESS), BM_GETCHECK, 0, 0) & BST_CHECKED ? TRUE : FALSE;
			else
				keepPress = FALSE;

			EndDialog(hWnd, 0);
			break;
		case IDCANCEL:
			SendMessage(hWnd, WM_CLOSE, 0, 0);	// tenuki
			break;
		case IDC_MOVERAISE:
			// MoveRaise -> KeepPress
			EnableWindow(GetDlgItem(hWnd, IDC_KEEPPRESS), SendMessage(GetDlgItem(hWnd, IDC_MOVERAISE), BM_GETCHECK, 0, 0) & BST_CHECKED);
			break;
		}
		break;
	case WM_CLOSE:
		EndDialog(hWnd, 1);
		break;
	}
	return FALSE;
}

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static char* backfile = NULL;
	static BOOL Cpressing = FALSE;
	static UINT Cx, Cy;
	static BOOL Lpressing = FALSE;
	static UINT Lx, Ly;

	switch(msg)
	{
	case WM_CREATE:
		if(!(field1 = malloc(WIDTH * HEIGHT * sizeof(int))))
		{
			MessageBox(NULL, "tB[hp1mێs", "G[", MB_ICONERROR | MB_OK);
			PostQuitMessage(0);
		}
		if(!(field2 = malloc(WIDTH * HEIGHT * sizeof(int))))
		{
			MessageBox(NULL, "tB[hp2mێs", "G[", MB_ICONERROR | MB_OK);
			PostQuitMessage(0);
		}
		if(!(field3 = malloc(WIDTH * HEIGHT * sizeof(int))))
		{
			MessageBox(NULL, "tB[hp3mێs", "G[", MB_ICONERROR | MB_OK);
			PostQuitMessage(0);
		}
		initwave();
		SetTimer(hWnd, 0, 20/*ms*/, NULL/*post WM_TIMER*/);
		break;
	case WM_SIZE:
		WIDTH = LOWORD(lParam);
		HEIGHT = HIWORD(lParam);
		free(field1);
		free(field2);
		free(field3);
		if(!(field1 = malloc(WIDTH * HEIGHT * sizeof(int))))
		{
			MessageBox(NULL, "tB[hp1mێs", "G[", MB_ICONERROR | MB_OK);
			PostQuitMessage(0);
		}
		if(!(field2 = malloc(WIDTH * HEIGHT * sizeof(int))))
		{
			MessageBox(NULL, "tB[hp2mێs", "G[", MB_ICONERROR | MB_OK);
			PostQuitMessage(0);
		}
		if(!(field3 = malloc(WIDTH * HEIGHT * sizeof(int))))
		{
			MessageBox(NULL, "tB[hp3mێs", "G[", MB_ICONERROR | MB_OK);
			PostQuitMessage(0);
		}
		break;
	case WM_CLOSE:
		PostQuitMessage(0);

		KillTimer(hWnd, 0);

		if(backfile)
			free(backfile);

		free(field1);
		free(field2);
		free(field3);

		break;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;

			hdc = BeginPaint(hWnd, &ps);
			if(hdc)	// G[N?
			{
				HBITMAP hbm;
				HDC mem;
				BITMAPINFO bmi;
				DWORD* bits;
				HBITMAP hbmold;

				ZeroMemory(&bmi, sizeof(BITMAPINFO));

				bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				bmi.bmiHeader.biBitCount = 32;
				bmi.bmiHeader.biPlanes = 1;
				bmi.bmiHeader.biWidth = WIDTH;
				bmi.bmiHeader.biHeight = -(int)HEIGHT;

				hbm = CreateDIBSection(NULL, &bmi, DIB_RGB_COLORS, &bits, NULL, 0);
				mem = CreateCompatibleDC(hdc);
				hbmold = SelectObject(mem, hbm);
				dowave(bits, backfile);
				BitBlt(hdc, 0, 0, WIDTH, HEIGHT, mem, 0, 0, SRCCOPY);
				SelectObject(mem, hbmold);
				DeleteObject(mem);
				DeleteObject(hbm);
			}
			EndPaint(hWnd, &ps);
		}
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case ID_CLEARFIELD:
			initwave();
			break;
		case ID_OPENBG:
			{
				OPENFILENAME ofn;
				char tmp[256];

				ZeroMemory(&ofn, sizeof(ofn));
				ofn.lStructSize = sizeof(ofn);
				ofn.hwndOwner = hWnd;
				ofn.hInstance = GetModuleHandle(NULL);
				ofn.lpstrFilter = "Bitmaps\0*.bmp\0";
				ofn.lpstrFile = tmp;
				tmp[0] = 0;
				ofn.nMaxFile = 256;
				ofn.lpstrTitle = "24bitBMPȊO͊J߂!!";
				ofn.Flags = OFN_FILEMUSTEXIST;
				if(GetOpenFileName(&ofn))
				{
					HANDLE fp;
					char* newbackfile;
					DWORD read;

					if(!(fp = CreateFile(tmp, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)))
					{
						MessageBox(hMainWnd, "t@Cǂݍݎs", "oops", MB_ICONERROR | MB_OK);
						break;
					}

					if(!(newbackfile = malloc(GetFileSize(fp, NULL))))
					{
						CloseHandle(fp);
						MessageBox(hMainWnd, "mێs", "oops", MB_ICONERROR | MB_OK);
						break;
					}

					ReadFile(fp, newbackfile, GetFileSize(fp, NULL), &read, NULL);

					CloseHandle(fp);

					if(backfile)
						free(backfile);
					backfile = newbackfile;
				}
			}
			break;
		case ID_ABOUT:
			{
				char name[MAX_PATH + 1];
				DWORD size;
				char* vi;

				GetModuleFileName(NULL, name, MAX_PATH);
				size = GetFileVersionInfoSize(name, 0);
				if(vi = HeapAlloc(GetProcessHeap(), 0, size))
				{
					GetFileVersionInfo(name, 0, size, vi);
					{
						VS_FIXEDFILEINFO* ver = (VS_FIXEDFILEINFO*)(vi + 40);

						// nameg
						wsprintf(name, "ripple by Murachue\n Version %d.%d.%d.%d", HIWORD(ver->dwFileVersionMS), LOWORD(ver->dwFileVersionMS),
															HIWORD(ver->dwFileVersionLS), LOWORD(ver->dwFileVersionLS));
						MessageBox(hWnd, name, "About ripple...", MB_ICONASTERISK | MB_OK);
					}
					HeapFree(GetProcessHeap(), 0, vi);
				}
			}
			break;
		case ID_CONFIGURE:
			DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CONFDIALOG), hWnd, ConfigureWndProc, 0);
			break;
		case ID_CLEARBG:
			if(backfile)
			{
				free(backfile);
				backfile = NULL;
			}
			break;
		}
		break;
	case WM_CONTEXTMENU:
		{
			HMENU hmp = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MAINMENU));
			HMENU hm = GetSubMenu(hmp, 0);
			TrackPopupMenuEx(hm, 0, LOWORD(lParam), HIWORD(lParam), hWnd, NULL);
			DestroyMenu(hm);
			DestroyMenu(hmp);
		}
		break;
	case WM_INITMENUPOPUP:	// j[o悤ƂĂ
		{
			int i, max = GetMenuItemCount((HMENU)wParam);
			MENUITEMINFO mii;

			for(i = 0; i < max; i++)
			{
				mii.cbSize = sizeof(mii);
				mii.fMask = MIIM_STATE;
				GetMenuItemInfo((HMENU)wParam, i, TRUE, &mii);

				switch(GetMenuItemID((HMENU)wParam, i))
				{
				case ID_CLEARBG:
					if(backfile)
						mii.fState &= ~MFS_DISABLED;	// L
					else
						mii.fState |= MFS_DISABLED;		// 
					break;
				}

				SetMenuItemInfo((HMENU)wParam, i, TRUE, &mii);
			}
		}
		break;
	case WM_MOUSEMOVE:
		if(Cpressing)
		{
			shiftx -= LOWORD(lParam) - Cx;
			shifty -= HIWORD(lParam) - Cy;
			Cx = LOWORD(lParam);
			Cy = HIWORD(lParam);
		}
		if(moveRaise && Lpressing)
		{
			Lx = LOWORD(lParam);
			Ly = HIWORD(lParam);
			if(!keepPress)
				raisewave(field2, Lx, Ly);
		}
		break;
	case WM_LBUTTONDOWN:
		Lx = LOWORD(lParam);
		Ly = HIWORD(lParam);
		raisewave(field2, Lx, Ly);
		Lpressing = TRUE;
		GetCapture();
		break;
	case WM_LBUTTONUP:
		Lpressing = FALSE;
		ReleaseCapture();
		break;
	case WM_MBUTTONDOWN:
		Cpressing = TRUE;
		Cx = LOWORD(lParam);
		Cy = HIWORD(lParam);
		GetCapture();
		break;
	case WM_MBUTTONUP:
		Cpressing = FALSE;
		ReleaseCapture();
		break;
	case WM_TIMER:
		if(keepPress && Lpressing)
			raisewave(field2, Lx, Ly);
		InvalidateRect(hWnd, NULL, FALSE);
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

int init(void)
{
	WNDCLASSEX wc;
	RECT rc;

	ZeroMemory(&wc, sizeof(wc));
	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)(COLOR_APPWORKSPACE + 1);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAINWNDICON));
	wc.lpfnWndProc = MainWndProc;
	wc.lpszClassName = "RIPPLEclass";
	if(!RegisterClassEx(&wc))
		return 1;

	// ClientWIDTHxHEIGHTɂ̂łB
	rc.left = rc.top = 0;
	rc.right = WIDTH;
	rc.bottom = HEIGHT;
	AdjustWindowRectEx(&rc, WS_OVERLAPPEDWINDOW & ~WS_SIZEBOX, FALSE, 0);
	rc.right -= rc.left;
	rc.bottom -= rc.top;

	if(!(hMainWnd = CreateWindowEx(0, wc.lpszClassName, "Ripple2: L=Raise M=Move R=Menu", WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT, rc.right, rc.bottom,
		NULL, NULL, wc.hInstance, NULL)))
		return 2;

	return 0;
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nShowCmd)
{
	MSG msg;

	if(init())
	{
		MessageBox(NULL, "s", "[", MB_ICONERROR | MB_OK);
		return 0;
	}

	ShowWindow(hMainWnd, SW_SHOW);

	while(GetMessage(&msg, NULL, 0, 0))	// Translate񂾂B
		DispatchMessage(&msg);

#ifdef NOCRT
	ExitProcess(msg.wParam);
#endif
	return 0;
}
